/////////////////////////////////////////////////
// BlueRoomElectronics UNICORN and 2x16 LCD
// Test / Exerciser / Learning and
// maybe even useful to someone besides me
//
// Author:
//  David Nienhiser -  Amateur Call: W9GF
//  Dallas, Texas USA
//  (Oval office for AMERICAS Team) :P
//
// Written for:
//   BlueRoomElectronics UNICORN + 2x16 LCD
//   I used my BlueRoomElectronics Junebug to program the UNICORN.
//
// Development Environment:
//
// 	MPLAB version 8.00
//  C Compiler - MPLAB C18 v3.16 (demo)
//  
//  Notes:
//   Memory Usage:
//		Program: 817 bytes, Data 272  (with demo/test code in main the program goes to 1931 bytes.
//
//   USEEIGHTDATABITS is a compile time constant - if it's set at TRUE this compiles for 8 bit LCD access
//				      if set to FALSE this compiles for 4 bit LCD access
//
//   This is my first PIC program - so you may need to cut me some slack!  I'm learning these things so
//   if you see ANYTHING that needs to be changed please let me know what it is - and why (cuz i'm learning).
//   I've attempted to document each of the functions.  Look at the code in main() - i believe i've called
//   every function here to exercise / test it.
//
//   I'm releasing this to the public domain!
//  

// Change Log
// -- When -----+------ Who -----------+----- What ------------------------+
// 08-Apr-2008		David Nienhiser			1.00: Created this monster!
// 13-Apr-2008		David Nienhiser			1.02: Adjustments for PowerTip PC1602K display (2x16)
//											While checking for busy bit in WaitLCDBusy() previously
//											was not toggling the Enable bit - changed this also had to
//											add a 1ms delay at the end of the WaitLCDBusy() code.
//											This didn't appear to be necessary for the Topway LMB162A display ?!?!
// 14-Apr-2008		David Nienhiser			1.03: Removed:	OSCCONbits.IOFS = TRUE;	// use internal oscillator 31.25khz internal clock
//											from the initialization code.

#include "p18f4550.h"

// Okay - here's these config things - i have no clue what they do - found one somewhere and it seemed to work fine
// then all of the sudden i couldn't debug in the IDE - then found another one and tried it and things seemed to work
// fine - then switched back and everything seemed fine also Grrrrrrrrrrrr!

#pragma config WDT = OFF, LVP = OFF, FOSC = HS

#include <stdlib.h>
#include <string.h>
#include <delays.h>

// Please don't change TRUE & FALSE!
#define	TRUE	1
#define	FALSE	0

// The following control what gets compiled.
#define	COMPILE_MAIN_AND_DEMO	TRUE	// TRUE = compile main() and demo code - FALSE = don't
#define	DELAYUSESTIMER0			TRUE	// TRUE = Use TIMER0 instead of 'speed bumps'
#define	USEEIGHTDATABITS		TRUE	// TRUE = compile for 8 data bits, FALSE = compile for 4 data bits

// Define types of LCD's Supported
#define	LCD2x16	1					// This is the only one i have (so far)!		
#define	LCD4x20	2					// Gonna have to get one of these & add to this!

#define	LCDTYPE	LCD2x16				// LCDTYPE controls conditional compilation options

#if		LCDTYPE == LCD2x16
 #define	LCDMINCOL	0
 #define	LCDMAXCOL	15
 #define	LCDMINROW	0
 #define	LCDMAXROW	1
 #define	LCDMIDDLECOLUMN	(((LCDMAXCOL-LCDMINCOL)/2)+1)
 #define	LCDMIDDLEROW  0		// Force to top row	((LCDMAXROW-LCDMINROW)/2)
#endif

#define INPUT	1
#define OUTPUT	0
#define BYTE	unsigned char
#define BOOL	unsigned char

#if		DELAYUSESTIMER0 == TRUE
 #define	TMR016BITVAL	64911			// with a 20mhz clock, 8:1 prescaler, 16bit timer this should result in a ET of .001 second		
#endif

#if		USEEIGHTDATABITS != TRUE
 #define	HIGHNIBBLEMASK	0xf0
 #define	LOWNIBBLEMASK	0x0f
#endif

// LCD Control Ports
//
// RS -  Register Select -  Port E Bit 0
//   0 (LCD_RS_COMMAND) -  	When writing to the LCD this is a LCD command
//						-	When reading from the LCD this is LCD status
//   1 (LCD_RS_DATA)	-   When writing to the LCD this is data to LCD
//						-   When reading from the LCD this is data from LCD
//
// The following constants and macros are for controlling the LCD's RS signal
#define LCD_RS_DATA		1
#define LCD_RS_COMMAND	0		
#define	LCD_RS	LATEbits.LATE0

// RW - Read/Write      -	Port E bit 1
//	0 (LCD_RW_WRITE)	-	Indicates data direction is from PIC to LCD (PIC->LCD)
//	1 (LCD_RW_READ)		-	Indicates data direction is from the LCD to the PIC (LCD->PIC)
//
// The following constants and macros are for controlling the LCD's RW signal
#define LCD_RW_READ		1
#define LCD_RW_WRITE	0
#define LCD_RW	LATEbits.LATE1

// E - Enable			-	Port E bit 2
//  1 to 0 (transition or edge)	-	Transfers data from PIC data register to LCD
//	0 to 1 (transition or edge) -	Transfers data from LCD to the PIC
//
// The following constants and macros are for controlling the LCD's E signal
#define LCD_E	LATEbits.LATE2

// LCD Busy Status bit -  Port D bit 7
//	1 (TRUE) after a read of the LCD DataPort with RS = LCD_RS_COMMAND and RW = LCD_RW_READ
//			 indicates that the LCD is busy.
//
// The LCD_BUSY macro defined below can bu used to access the LCD Busy bit.
#define	LCD_BUSY	PORTDbits.RD7

// The LCD_DATA_PORT macro is used to access the LCD Data Port
#define	LCD_DATA_PORT	LATD

// The LCD_BL macro is used to control the LCD backlight
// LCD_BL = TRUE enables the backlight
// LCD_BL = FALSE disables the backlight
#define	LCD_BL		LATBbits.LATB3

// LCD command codes defined here
#define	LCD_CMD_CLEARDISPLAY		0x01	//Clear Display (only option in CMD 0)
#define	LCD_CMD_DISPLAYHOMECURSOR	0x02	//Display and home cursor (only option in CMD1)
#define	LCD_CMD2BIT					0x04	//Command 2 (Character Entry Mode) has bit 2 set
 #define	LCD_CMD2BIT_INC_DEC		0x02	//1=Increment, 0=Decrement
 #define LCD_CMD2BIT_DISPSHIFT		0x01	//1=Display Shift On, 0=Display Shift Off
 #define	LCDCMD2BITS		LCD_CMD2BIT | LCD_CMD2BIT_INC_DEC	// Command 2 to use here.
#define	LCDCMD3BIT					0x08	//Command 3 (Display on/off & Cursor) has bit 3 set
 #define LCDCMD3BIT_DISPLAY_ON	 	0x04	//Display ON bit 
 #define LCDCMD3BIT_UNDERSCORE	 	0x02	//1 = Cursor underline ON, 0 = OFF
 #define LCDCMD3BIT_CURSOR_BLINK 	0x01	//1 = Cursor Blink ON, 0 = OFF
 #define LCDCMD3BITS	LCDCMD3BIT | LCDCMD3BIT_DISPLAY_ON	//Command 3 to use here.
#define	LCDCMD4BIT					0x80	//Command 4 (Display/Cursor Shift) has bit 4 set
 #define LCDCMD4BIT_DISPSHIFT		0x08	//1=Display Shift, 0=Cursor Move
 #define LCDCMD4BIT_RLSHIFT			0x04	//1=Right Shift, 0=Left Shift
#define	LCDCMD5BIT					0x20	//Command 5 (Function Set) has bit 5 set
 #define	LCDCMD5BIT_8BIT			0x10	//1=8 bit interface, 0=4bit interface
 #define	LCDCMD5BIT_2LINES		0x08	//1=2 line mode, 0=1 line mode 
 #define	LCDCMD5BIT_10DOTFMT		0x04	//1=5x10 dot format, 0=5x7 dot format
 #if USEEIGHTDATABITS == TRUE
  #define	LCDCMD5BITS		LCDCMD5BIT|LCDCMD5BIT_8BIT|LCDCMD5BIT_2LINES|LCDCMD5BIT_10DOTFMT //Command 5 to use here
 #else
  #define	LCDCMD5BITS		LCDCMD5BIT|LCDCMD5BIT_2LINES|LCDCMD5BIT_10DOTFMT //Command 5 to use here 
  #define	LCDCMD5BITSFIRST LCDCMD5BIT | LCDCMD5BIT_8BIT  // This gets sent first in 4 bit mode
 #endif 
#define	LCDCMD6BIT					0x40	//Command 6 (Set CGRAM Address) has bit 6 set, bits 0-5 = address
#define	LCDCMD7BIT					0x80	//Command 7 (Set Display Address) has bit 7 set, bits 0-6 = address

// Function prototypes to keep the compiler happy :)
void InitializeLCD(BOOL Backlight);
void Delay(int milliseconds);
void ControlLCDBacklight(BOOL State);
void WaitifLCDBusy(void);
void SendLCDCharacter(unsigned char character);
void SendLCDCommand(BYTE command, BOOL waitLCDReady);
void SetLCDCursor(BYTE Row, BYTE Column);
BYTE SendROMStringToLCD(const rom char *zString);
BYTE SendRAMStringToLCD(ram char *zString);
BYTE CenterAndSendRAMStringToLCD(BYTE Row, ram char*zString);
BYTE CenterAndSendROMStringToLCD(BYTE Row, const rom char*zString);
BYTE DisplayInteger(BYTE Row, BYTE Column, int value);
void LoadUserDefinedLCDCharacter(BYTE StartingCGRamAddress,BYTE NumberOfCharacters, const rom char* CharacterDefinition);
void LoadUserDefinedLCDCharacter(BYTE StartingCGRamAddress,BYTE NumberOfCharacters, const rom char* CharacterDefinition);

// Useful Shortcut(s) (IMHO)
#define LCDClearScreen() 	SendLCDCommand(LCD_CMD_CLEARDISPLAY,TRUE)



#if COMPILE_MAIN_AND_DEMO == TRUE
void main(void)
{
	char dataarray[]="W9GF Unicorn&LCD";
	char dataarray2[]="Centered";
	int counter,counter2,counter3,counter4,row,col,vardelay;

	InitializeLCD(TRUE);
	
	#define SECOND		1000
	#define TESTDELAY	3*SECOND
	
	while(TRUE==TRUE) {
		LCDClearScreen();
		SendRAMStringToLCD(dataarray);
		SetLCDCursor(LCDMAXROW,LCDMINCOL);
		#if USEEIGHTDATABITS == FALSE
		 SendROMStringToLCD("  Demo (4 Bit)  ");
		#else
		 SendROMStringToLCD("  Demo (8 Bit)  ");
		#endif				
		for(counter=LCDMAXCOL;counter>=LCDMINCOL;counter--) {
			SetLCDCursor(LCDMINROW,counter);
			SendLCDCharacter(dataarray[counter]);
		}
		Delay(TESTDELAY);
		LCDClearScreen();
		CenterAndSendROMStringToLCD(LCDMINROW,"Written by");
		CenterAndSendROMStringToLCD(LCDMINROW+1,"David Nienhiser");
		Delay(TESTDELAY);
		LCDClearScreen();
		CenterAndSendROMStringToLCD(LCDMINROW,"Version 1.03");
		CenterAndSendROMStringToLCD(LCDMINROW+1,"14-Apr-2008");
		Delay(TESTDELAY);
		LCDClearScreen();
		CenterAndSendROMStringToLCD(LCDMINROW,"Recompile");
		#if USEEIGHTDATABITS == TRUE
		  CenterAndSendROMStringToLCD(LCDMINROW+1,"for 4 data bits!");
		#else
		  CenterAndSendROMStringToLCD(LCDMINROW+1,"for 8 data bits!");
		#endif
		Delay(TESTDELAY);
		// Code to test centering
		LCDClearScreen();
		SendROMStringToLCD("1234567887654321");	
		CenterAndSendRAMStringToLCD(1,dataarray2);
		Delay(TESTDELAY);
		LCDClearScreen();	
		SendROMStringToLCD("1234567887654321");
		CenterAndSendROMStringToLCD(1,"||Centered||");
		Delay(TESTDELAY);
		LCDClearScreen();
		SetLCDCursor(LCDMINROW,LCDMINCOL);
		SendROMStringToLCD("X=");
		SetLCDCursor(LCDMINROW,LCDMIDDLECOLUMN);
		SendROMStringToLCD("Y=");
		SetLCDCursor(LCDMINROW+1,LCDMINCOL);
		SendROMStringToLCD("Z=");
		SetLCDCursor(LCDMINROW+1,LCDMIDDLECOLUMN);
		SendROMStringToLCD("$=");
		for(counter=0,counter2=100,counter3=1000,counter4=-20000;counter<500;
		    ++counter,counter2+=10,counter3+=15,counter4+=500) {
			DisplayInteger(LCDMINROW,LCDMINCOL+2,counter);	    
			DisplayInteger(LCDMINROW,LCDMIDDLECOLUMN+4,counter2);
			DisplayInteger(LCDMINROW+1,LCDMINCOL+2,counter3);
			DisplayInteger(LCDMINROW+1,LCDMIDDLECOLUMN+2,counter4);	
			Delay(10);	    
		}
		LCDClearScreen();
		// Load a Stick Man as character 0
		LoadUserDefinedLCDCharacter(0,1,"\x0e\x11\x0e\x04\x1f\x04\x0a\x11");		// Stick Man			
		for(vardelay=600,row=LCDMINROW,col=LCDMINCOL;col<=LCDMAXCOL;++col,vardelay-=15) {
			if(col>LCDMINCOL) {		// erase where he was	
				SetLCDCursor(row,col-1);
				SendLCDCharacter(' ');
			}
			SetLCDCursor(row,col);
			SendLCDCharacter(0);
			Delay(vardelay);
		}
		SetLCDCursor(LCDMINROW,LCDMAXCOL);
		SendLCDCharacter(' ');
		for(row=LCDMINROW+1,col=LCDMAXCOL;col>=LCDMINCOL;--col,vardelay-=15) {
			if(col<LCDMAXCOL) {
				SetLCDCursor(row,col+1);
				SendLCDCharacter(' ');
			}
			SetLCDCursor(row,col);
			SendLCDCharacter(0);
			Delay(vardelay);
			
		}
		Delay(SECOND/2);
		LCDClearScreen();
		for(counter=LCDMINCOL;counter<=LCDMAXCOL;++counter)
			SendLCDCharacter(0);
		SetLCDCursor(LCDMINROW+1,LCDMINCOL);
		for(counter=LCDMINCOL;counter<=LCDMAXCOL;++counter)
			SendLCDCharacter(0);
		Delay(TESTDELAY);				

		// Load a backslash (\) as character 7
		LoadUserDefinedLCDCharacter(7,1,"\x20\x10\x08\x04\x02\x01\x0\x0");			// Backslash (\)
		LCDClearScreen();
		SendROMStringToLCD("We have forward");
		SetLCDCursor(LCDMINROW+1,LCDMINCOL);
		SendROMStringToLCD("slashes: ////");
		Delay(TESTDELAY);
		LCDClearScreen();
		SendROMStringToLCD("We have back-");
		SetLCDCursor(LCDMINROW+1,LCDMINCOL);
		SendROMStringToLCD("slashes: \x7\x7\x7\x7");
		Delay(TESTDELAY);
		LCDClearScreen();
		CenterAndSendROMStringToLCD(LCDMINROW,"We just LOVE");
		CenterAndSendROMStringToLCD(LCDMINROW+1,"our UNICORN!!");
		for(counter=19;counter;--counter) {
			ControlLCDBacklight(counter&1);
			Delay(SECOND/4);
		}
		ControlLCDBacklight(TRUE);
		Delay(TESTDELAY);
	}		
}
#endif

// void LoadUserDefinedLCDCharacter(BYTE StartingCGRamAddress,BYTE NumberOfCharacters, const rom char* CharacterDefinition)
//	StartingCGRamAddress -  Starting CHARACTER position for load
//  NumberofCharacters   -  Number of CHARACTERS to load (not number of bytes)
//  CharacterDefinition  -  Points to a character definition table - 8 per character - this does NOT look at NULL termination!
//
// This function loads a user defined bitmap into the LCD character generator.
//
void LoadUserDefinedLCDCharacter(BYTE StartingCGRamAddress,BYTE NumberOfCharacters, const rom char* CharacterDefinition)
{
	BYTE temp1,temp2;
	
		SendLCDCommand((StartingCGRamAddress*8)|LCDCMD6BIT,TRUE);		
		for(NumberOfCharacters*=8;NumberOfCharacters;--NumberOfCharacters,++CharacterDefinition)		
			SendLCDCharacter(*CharacterDefinition);
}

// BYTE DisplayInteger(BYTE Row, BYTE Column, int value)
//	Row -  Row number to display data on
//  Column - Starting Column number to display data on
//  value -  32bit signed value to display
//
// Returns: 8 bit unsigned value that is the length of the string written.
//
// This function converts a 16 bit signed value (integer) to a string then displays
// it on the LCD at the Row/Column values specified in the
// parameter.
//
BYTE DisplayInteger(BYTE Row, BYTE Column, int value)
{
	char conversionbuffer[12];
	
	SetLCDCursor(Row,Column);
	return(SendRAMStringToLCD(itoa(value,conversionbuffer)));	
}

// BYTE CenterAndSendRAMStringToLCD(BYTE Row, ram char*zString)
//	Row -  Row number to display on
//  zString - Points to NULL terminated string in RAM memory.
//
// Returns: 8 bit unsigned value that is the length of the string written.
//
// This function center's a string on the line specified in the parameter.
// The string must reside in RAM
// 
BYTE CenterAndSendRAMStringToLCD(BYTE Row, ram char*zString)
{
	SetLCDCursor(Row,LCDMIDDLECOLUMN-(strlen(zString)/2));
	return(SendRAMStringToLCD(zString));
}

// BYTE SendRAMStringToLCD(ram char *zString)
//	zString = pointer to a NULL terminated string in RAM memory
//
// Returns: 8 bit unsigned value that is the length of the string written.
//
// Send NULL terminated string (zString) residing in RAM to LCD
// starting at current cursor location.  MPLAB C18 stores strings &
// variables in RAM memory.
//
// Return: void
//
BYTE SendRAMStringToLCD(ram char *zString)
{
	BYTE bCounter=0;
	
	while(*zString) {
		++bCounter;
		SendLCDCharacter(*zString++);
	}
	return(bCounter);
}

// BYTE CenterAndSendROMStringToLCD(BYTE Row, const rom char*zString)
//	Row -  Row number to display on
//  zString - Points to NULL terminated string in ROM memory.
//
// Returns: 8 bit unsigned value that is the length of the string written.
//
// This function center's a string on the line specified in the parameter.
// The string must reside in ROM
// 
BYTE CenterAndSendROMStringToLCD(BYTE Row, const rom char* zString)
{
	BYTE stringlength;
	
	for(stringlength=0;zString[stringlength]!='\0';++stringlength)
		;
	SetLCDCursor(Row,LCDMIDDLECOLUMN-(stringlength/2));
	return(SendROMStringToLCD(zString));
}

// BYTE SendROMStringToLCD(const rom char *zString)
//	zString = pointer to a NULL terminated string in program memory
//
// Returns: 8 bit unsigned value that is the length of the string written.
//
// Send NULL terminated string (zString) residing in RAM to LCD
// starting at current cursor location.  MPLAB C18 stores constant
// literal variables & strings in program memory.
//
BYTE SendROMStringToLCD(const rom char *zString)
{
	BYTE bCounter=0;
	while(*zString) {
		++bCounter;
		SendLCDCharacter(*zString++);
	}
	return(bCounter);
}
	
// void SetLCDCursor(BYTE Row, BYTE Column)
//	Row    = Row Number
//	Column = Column Number
//
// Position LCD cursor at Row, Column
//
void SetLCDCursor(BYTE Row, BYTE Column)
{
	SendLCDCommand((Column&0x0f)+(Row?0x40:0)|LCDCMD7BIT,TRUE);
}		

// void InitializeLCD(BOOL Backlight)
//	Backlight = TRUE to enable the backlight after initialization
//			  = FALSE to leave the backlight turned off
//  This function should only be called once in the beginning - it's purpose is to
//  configure the PIC ports to talk to the LCD and to initialize the LCD.
//  Once the Command 5 byte has been sent to the LCD subsequent calls to send
//  initialization commands will wait for the LCD Busy status before sending
//  commands to the LCD.
void InitializeLCD(BOOL Backlight)
{
	int SoftResetCounter;

	#if DELAYUSESTIMER0 == TRUE
	  // Get TIMER0 setup for use in Delay() code below
	  // Use: 1:32 Prescale - this should give a time period of .0016 seconds
	  // Close enough for government work!
	 T0CONbits.TMR0ON=FALSE;				// Just in case we will stop it while we set it up.
	 T0CONbits.T08BIT=FALSE;				// Setup for 16 bits (1 = 8 bit, 0 = 16) 
	 TMR0L=0;
	 T0CONbits.PSA=0;					// 0 = Use Prescaler, 1= do not
	 T0CONbits.T0CS=FALSE;				// 0 = Use internal instruction clock
	 T0CONbits.T0PS0=T0CONbits.T0PS2=FALSE; 	//1:8 prescale value
	 T0CONbits.T0PS1=TRUE;
	 
	#endif
		
	ADCON1 = 0xff;		// All I/O = Digital?
	// Setup LCD Backlight Control pin
	#if USEEIGHTDATABITS == TRUE
	  TRISD = OUTPUT;		// Initial setup for LCD Data Port = output
	#else
	  TRISD = 0x0f;			// Just use upper 4 if 4 bit mode.
	#endif  
	LCD_BL = FALSE;		// Explicitly turn backlight off.
	TRISBbits.TRISB3=OUTPUT;
	PORTE = 0;			// Clear PORTE output data
	TRISE = 0;
	LATE = 0;			// Immediately after this statement is where the OSCCONbits.IOFS=TRUE was
						// Removed 14Apr2008
//	OSCCONbits.IOFS = TRUE;	// use internal oscillator 31.25khz internal clock
	// LCD Soft reset (ignores busy flag)
	LCD_RS = LCD_RS_COMMAND;  // Indicate to LCD we are sending command
	LCD_RW = LCD_RW_WRITE;	  // Tell LCD we are writing
	#if	USEEIGHTDATABITS == FALSE
	  SendLCDCommand(LCDCMD5BITSFIRST,FALSE);
	  Delay(60);			
	  SendLCDCommand(LCDCMD5BITSFIRST,FALSE);
	  Delay(60);
	#endif
	for(SoftResetCounter=0;SoftResetCounter<4;++SoftResetCounter) {
		#if	USEEIGHTDATABITS == TRUE
  		  Delay(60);
  		  SendLCDCommand(LCDCMD5BITS,FALSE);
		#else 
		  Delay(60);
		  SendLCDCommand(LCDCMD5BITS,FALSE);
		#endif			
	}
	SendLCDCommand(LCDCMD3BITS,TRUE);	
	SendLCDCommand(LCD_CMD_CLEARDISPLAY,TRUE);
	SendLCDCommand(LCDCMD2BITS,TRUE);
	if(Backlight)
		ControlLCDBacklight(TRUE);
}

// SendLCDCommand(BYTE command, BOOL waitLCDReady)
//	command = command to be sent to LCD
//  waitLCDReady -	TRUE = Wait for READY from LCD before writing command
//					FALSE = Do not wait for READY from LCD before writing command
//
// This function sends a command to the LCD - before sending it checks the LCD Busy status
// and if the LCD is busy it waits until it's not (maybe forever)
//
void SendLCDCommand(BYTE command, BOOL waitLCDReady)
{
	if(waitLCDReady)
		WaitifLCDBusy();	// Check if busy and wait till ready
	LCD_RS = LCD_RS_COMMAND;  // Indicate to LCD we are sending command
	LCD_RW = LCD_RW_WRITE;	  // Tell LCD we are writing
	#if	USEEIGHTDATABITS == TRUE
	  LCD_E = TRUE;	
	  LCD_DATA_PORT = command;		
	  LCD_E = FALSE;	
	#else
	  LCD_E = TRUE;	
	  LCD_DATA_PORT = (command&HIGHNIBBLEMASK)|(LCD_DATA_PORT&LOWNIBBLEMASK);		
	  LCD_E = FALSE;
	  LCD_E = TRUE;	  
	  LCD_DATA_PORT = ((command&LOWNIBBLEMASK)<<4)|(LCD_DATA_PORT&LOWNIBBLEMASK) ;		
	  LCD_E = FALSE;		
	#endif
Delay(1);
}	

// void SendLCDCharacter(unsigned char character)
//	character = character to be sent to the LCD
//
// This function sends the character passed as the parameter to the LCD at the current
// cursor position.  Before sending the character this function waits for the LCD Busy status
// and if the LCD is busy it waits until it's not (maybe forever)
void SendLCDCharacter(unsigned char character)
{
	WaitifLCDBusy();
	LCD_RS = LCD_RS_DATA;
	LCD_RW = LCD_RW_WRITE;
	#if	USEEIGHTDATABITS == TRUE
	  LCD_E = TRUE;	
	  LCD_DATA_PORT = character;
	  LCD_E = FALSE;			
	#else
	  LCD_E = TRUE;	
	  LCD_DATA_PORT = (character&HIGHNIBBLEMASK)|(LCD_DATA_PORT&LOWNIBBLEMASK);		
	  LCD_E = FALSE;
	  LCD_E = TRUE;	  
	  LCD_DATA_PORT = ((character&LOWNIBBLEMASK)<<4)|(LCD_DATA_PORT&LOWNIBBLEMASK);				
	  LCD_E = FALSE;					  	  
	#endif	
}	

// void WaitifLCDBusy(void)
//
// This function checks the LCD busy status and if busy waits until it's ready.
//
void WaitifLCDBusy(void)
{
	TRISDbits.TRISD7 = INPUT;
	LCD_RS = LCD_RS_COMMAND;
	LCD_RW = LCD_RW_READ;
	LCD_E = TRUE;
	while(LCD_BUSY == TRUE) {
		LCD_E=FALSE;
		LCD_E=TRUE;
		
	}
	LCD_E = FALSE;
	LCD_RW = LCD_RW_WRITE;	
	TRISDbits.TRISD7 = OUTPUT;	
	Delay(1);
}

// void ControlLCDBacklight(BOOL State)
//	State =	TRUE turns the LCD backlight ON
//		  = FALSE turns the LCD backlight OFF
//	
void ControlLCDBacklight(BOOL State)
{
	LCD_BL=State==TRUE?TRUE:FALSE;
}

// void Delay(int milliseconds)
// 	milliseconds -  number of milliseconds to delay
//
// This functions introduces "speed bumps" in the code - there's got to be a BETTER WAY
// like using a timer?  I'll work on that - bigger fish to fry at the moment though!  I think
// i'll put a compile conditional in the beginning as a memory jogger!
//
void Delay(int milliseconds)
{
	
	#if	DELAYUSESTIMER0 == TRUE
	 milliseconds=milliseconds>5*SECOND?5*SECOND:milliseconds;	// no reason to delay more than 5 seconds!	 
	 TMR0H=(TMR016BITVAL>>8)&0x00ff;		
	 TMR0L=TMR016BITVAL&0x00ff;				// Seed the timer values
	 T0CONbits.TMR0ON=TRUE;					// Turn 'em on (GIT-R-DONE!)
	 while(milliseconds) {
			if(INTCONbits.T0IF==TRUE) {
				TMR0H=(TMR016BITVAL>>8)&0x00ff;			
	 			TMR0L=TMR016BITVAL&0x00ff;				// Seed the timer values
	 			INTCONbits.T0IF=FALSE;
				--milliseconds;
			}
	 }
	 T0CONbits.TMR0ON=FALSE;				// Turn 'em off
	#else
	int outer,inner;
	milliseconds=milliseconds>5*SECOND?5*SECOND:milliseconds;	// no reason to delay more than 5 seconds!
 	 for(outer=0;outer<milliseconds;outer++)
		for(inner=0;inner<275;inner++)
			;
	#endif
}
			